#!/usr/bin/env python

"""
Python interface to CULA toolkit.
"""

from __future__ import absolute_import

import sys
import ctypes
import atexit
import numpy as np

from . import cuda

# Load CULA library:
if 'linux' in sys.platform:
    _libcula_libname_list = ['libcula_lapack.so',
                             'libcula_lapack_basic.so',
                             'libcula.so']
elif sys.platform == 'darwin':
    _libcula_libname_list = ['libcula_lapack.dylib',
                             'libcula.dylib']
elif sys.platform == 'win32':
    _libcula_libname_list = ['cula_lapack.dll',
                             'cula_lapack_basic.dll']
else:
    raise RuntimeError('unsupported platform')

_load_err = ''
for _lib in  _libcula_libname_list:
    try:
        _libcula = ctypes.cdll.LoadLibrary(_lib)
    except OSError:
        _load_err += ('' if _load_err == '' else ', ') + _lib
    else:
        _load_err = ''
        break
if _load_err:
    raise OSError('%s not found' % _load_err)

# Check whether the free or standard version of the toolkit is
# installed by trying to access a function that is only available in
# the latter:
try:
    _libcula.culaDeviceMalloc
except AttributeError:
    _libcula_toolkit = 'free'
else:
    _libcula_toolkit = 'standard'

# Function for retrieving string associated with specific CULA error
# code:
_libcula.culaGetStatusString.restype = ctypes.c_char_p
_libcula.culaGetStatusString.argtypes = [ctypes.c_int]
def culaGetStatusString(e):
    """
    Get string associated with the specified CULA status code.

    Parameters
    ----------
    e : int
        Status code.

    Returns
    -------
    s : str
        Status string.

    """

    return _libcula.culaGetStatusString(e)

# Generic CULA error:
class culaError(Exception):
    """CULA error."""
    pass

# Exceptions corresponding to various CULA errors:
class culaNotFound(culaError):
    """CULA shared library not found"""
    pass

class culaStandardNotFound(culaError):
    """Standard CULA Dense toolkit unavailable"""
    pass

class culaNotInitialized(culaError):
    try:
        __doc__ = culaGetStatusString(1)
    except:
        pass
    pass

class culaNoHardware(culaError):
    try:
        __doc__ = culaGetStatusString(2)
    except:
        pass
    pass

class culaInsufficientRuntime(culaError):
    try:
        __doc__ = culaGetStatusString(3)
    except:
        pass
    pass

class culaInsufficientComputeCapability(culaError):
    try:
        __doc__ = culaGetStatusString(4)
    except:
        pass
    pass

class culaInsufficientMemory(culaError):
    try:
        __doc__ = culaGetStatusString(5)
    except:
        pass
    pass

class culaFeatureNotImplemented(culaError):
    try:
        __doc__ = culaGetStatusString(6)
    except:
        pass
    pass

class culaArgumentError(culaError):
    try:
        __doc__ = culaGetStatusString(7)
    except:
        pass
    pass

class culaDataError(culaError):
    try:
        __doc__ = culaGetStatusString(8)
    except:
        pass
    pass

class culaBlasError(culaError):
    try:
        __doc__ = culaGetStatusString(9)
    except:
        pass
    pass

class culaRuntimeError(culaError):
    try:
        __doc__ = culaGetStatusString(10)
    except:
        pass
    pass

class culaBadStorageFormat(culaError):
    try:
        __doc__ = culaGetStatusString(11)
    except:
        pass
    pass

class culaInvalidReferenceHandle(culaError):
    try:
        __doc__ = culaGetStatusString(12)
    except:
        pass
    pass

class culaUnspecifiedError(culaError):
    try:
        __doc__ = culaGetStatusString(13)
    except:
        pass
    pass

culaExceptions = {
    -1: culaNotFound,
    1: culaNotInitialized,
    2: culaNoHardware,
    3: culaInsufficientRuntime,
    4: culaInsufficientComputeCapability,
    5: culaInsufficientMemory,
    6: culaFeatureNotImplemented,
    7: culaArgumentError,
    8: culaDataError,
    9: culaBlasError,
    10: culaRuntimeError,
    11: culaBadStorageFormat,
    12: culaInvalidReferenceHandle,
    13: culaUnspecifiedError,
    }

# CULA functions:
_libcula.culaGetErrorInfo.restype = int
def culaGetErrorInfo():
    """
    Returns extended information code for the last CULA error.

    Returns
    -------
    err : int
        Extended information code.
    """

    return _libcula.culaGetErrorInfo()

_libcula.culaGetErrorInfoString.restype = int
_libcula.culaGetErrorInfoString.argtypes = [ctypes.c_int,
                                            ctypes.c_void_p,
                                            ctypes.c_void_p,
                                            ctypes.c_int]
def culaGetErrorInfoString(e, i, bufsize=100):
    """
    Returns a readable CULA error string.

    Returns a readable error string corresponding to a given CULA
    error code and extended error information code.

    Parameters
    ----------
    e : int
        CULA error code.
    i : int
        Extended information code.
    bufsize : int
        Length of string to return.

    Returns
    -------
    s : str
        Error string.
    """

    buf = ctypes.create_string_buffer(bufsize)
    status = _libcula.culaGetErrorInfoString(e, i, buf, bufsize)
    culaCheckStatus(status)
    return buf.value

def culaGetLastStatus():
    """
    Returns the last status code returned from a CULA function.

    Returns
    -------
    s : int
        Status code.
    """

    return _libcula.culaGetLastStatus()

def culaCheckStatus(status):
    """
    Raise an exception corresponding to the specified CULA status
    code.

    Parameters
    ----------
    status : int
        CULA status code.
    """

    if status != 0:
        error = culaGetErrorInfo()
        try:
            e = culaExceptions[status](error)
        except KeyError:
            raise culaError(error)
        else:
            raise e

_libcula.culaSelectDevice.restype = int
_libcula.culaSelectDevice.argtypes = [ctypes.c_int]
def culaSelectDevice(dev):
    """
    Selects a device with which CULA will operate.

    Parameters
    ----------
    dev : int
        GPU device number.

    Notes
    -----
    Must be called before `culaInitialize`.
    """

    status = _libcula.culaSelectDevice(dev)
    culaCheckStatus(status)

_libcula.culaGetExecutingDevice.restype = int
_libcula.culaGetExecutingDevice.argtypes = [ctypes.c_void_p]
def culaGetExecutingDevice():
    """
    Reports the id of the GPU device used by CULA.

    Returns
    -------
    dev : int
       Device id.
    """

    dev = ctypes.c_int()
    status = _libcula.culaGetExecutingDevice(ctypes.byref(dev))
    culaCheckStatus(status)
    return dev.value

def culaFreeBuffers():
    """
    Releases any memory buffers stored internally by CULA.
    """

    _libcula.culaFreeBuffers()

_libcula.culaGetVersion.restype = int
def culaGetVersion():
    """
    Report the version number of CULA.
    """

    return _libcula.culaGetVersion()

_libcula.culaGetCudaMinimumVersion.restype = int
def culaGetCudaMinimumVersion():
    """
    Report the minimum version of CUDA required by CULA.
    """

    return _libcula.culaGetCudaMinimumVersion()

_libcula.culaGetCudaRuntimeVersion.restype = int
def culaGetCudaRuntimeVersion():
    """
    Report the version of the CUDA runtime linked to by the CULA library.
    """

    return _libcula.culaGetCudaRuntimeVersion()

_libcula.culaGetCudaDriverVersion.restype = int
def culaGetCudaDriverVersion():
    """
    Report the version of the CUDA driver installed on the system.
    """

    return _libcula.culaGetCudaDriverVersion()

_libcula.culaGetCublasMinimumVersion.restype = int
def culaGetCublasMinimumVersion():
    """
    Report the version of CUBLAS required by CULA.
    """

    return _libcula.culaGetCublasMinimumVersion()

_libcula.culaGetCublasRuntimeVersion.restype = int
def culaGetCublasRuntimeVersion():
    """
    Report the version of CUBLAS linked to by CULA.
    """

    return _libcula.culaGetCublasRuntimeVersion()

_libcula.culaGetDeviceCount.restype = int
def culaGetDeviceCount():
    """
    Report the number of available GPU devices.
    """

    return _libcula.culaGetDeviceCount()

_libcula.culaInitialize.restype = int
def culaInitialize():
    """
    Initialize CULA.

    Notes
    -----
    Must be called before using any other CULA functions.
    """

    status = _libcula.culaInitialize()
    culaCheckStatus(status)

_libcula.culaShutdown.restype = int
def culaShutdown():
    """
    Shuts down CULA.
    """

    status = _libcula.culaShutdown()
    culaCheckStatus(status)

# Shut down CULA upon exit:
atexit.register(_libcula.culaShutdown)

# LAPACK functions available in CULA Dense Free:

# SGESV, CGESV
_libcula.culaDeviceSgesv.restype = \
_libcula.culaDeviceCgesv.restype = int
_libcula.culaDeviceSgesv.argtypes = \
_libcula.culaDeviceCgesv.argtypes = [ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_void_p,
                                     ctypes.c_int]
def culaDeviceSgesv(n, nrhs, a, lda, ipiv, b, ldb):
    """
    Solve linear system with LU factorization.
    """

    status = _libcula.culaDeviceSgesv(n, nrhs, int(a), lda, int(ipiv),
                                      int(b), ldb)
    culaCheckStatus(status)

def culaDeviceCgesv(n, nrhs, a, lda, ipiv, b, ldb):
    """
    Solve linear system with LU factorization.
    """

    status = _libcula.culaDeviceCgesv(n, nrhs, int(a), lda, int(ipiv),
                                      int(b), ldb)
    culaCheckStatus(status)

# SGETRF, CGETRF
_libcula.culaDeviceSgetrf.restype = \
_libcula.culaDeviceCgetrf.restype = int
_libcula.culaDeviceSgetrf.argtypes = \
_libcula.culaDeviceCgetrf.argtypes = [ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def culaDeviceSgetrf(m, n, a, lda, ipiv):
    """
    LU factorization.
    """

    status = _libcula.culaDeviceSgetrf(m, n, int(a), lda, int(ipiv))
    culaCheckStatus(status)

def culaDeviceCgetrf(m, n, a, lda, ipiv):
    """
    LU factorization.
    """

    status = _libcula.culaDeviceCgetrf(m, n, int(a), lda, int(ipiv))
    culaCheckStatus(status)

# SGEQRF, CGEQRF
_libcula.culaDeviceSgeqrf.restype = \
_libcula.culaDeviceCgeqrf.restype = int
_libcula.culaDeviceSgeqrf.argtypes = \
_libcula.culaDeviceCgeqrf.argtypes = [ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def culaDeviceSgeqrf(m, n, a, lda, tau):
    """
    QR factorization.
    """

    status = _libcula.culaDeviceSgeqrf(m, n, int(a), lda, int(tau))
    culaCheckStatus(status)

def culaDeviceCgeqrf(m, n, a, lda, tau):
    """
    QR factorization.
    """

    status = _libcula.culaDeviceCgeqrf(m, n, int(a), lda, int(tau))
    culaCheckStatus(status)

# SORGQR, CUNGQR
try:
    _libcula.culaDeviceSorgqr.restype = \
    _libcula.culaDeviceCungqr.restype = int
    _libcula.culaDeviceSorgqr.argtypes = \
    _libcula.culaDeviceCungqr.argtypes = [ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p]
except AttributeError:
    def culaDeviceSorgqr(m, n, k, a, lda, tau):
        """
        QR factorization - Generate Q from QR factorization
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCungqr(m, n, k, a, lda, tau):
        """
        QR factorization - Generate Q from QR factorization
        """
        
        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSorgqr(m, n, k, a, lda, tau):
        """
        QR factorization - Generate Q from QR factorization
        """

        status = _libcula.culaDeviceSorgqr(m, n, k, int(a), lda, int(tau))
        culaCheckStatus(status)

    def culaDeviceCungqr(m, n, k, a, lda, tau):
        """
        QR factorization - Generate Q from QR factorization
        """

        status = _libcula.culaDeviceCungqr(m, n, k, int(a), lda, int(tau))
        culaCheckStatus(status)

# SGELS, CGELS
_libcula.culaDeviceSgels.restype = \
_libcula.culaDeviceCgels.restype = int
_libcula.culaDeviceSgels.argtypes = \
_libcula.culaDeviceCgels.argtypes = [ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int]
def culaDeviceSgels(trans, m, n, nrhs, a, lda, b, ldb):
    """
    Solve linear system with QR or LQ factorization.
    """

    trans = trans.encode('ascii')
    status = _libcula.culaDeviceSgels(trans, m, n, nrhs, int(a),
                                      lda, int(b), ldb)
    culaCheckStatus(status)

def culaDeviceCgels(trans, m, n, nrhs, a, lda, b, ldb):
    """
    Solve linear system with QR or LQ factorization.
    """

    trans = trans.encode('ascii')
    status = _libcula.culaDeviceCgels(trans, m, n, nrhs, int(a),
                                      lda, int(b), ldb)
    culaCheckStatus(status)

# SGGLSE, CGGLSE
_libcula.culaDeviceSgglse.restype = \
_libcula.culaDeviceCgglse.restype = int
_libcula.culaDeviceSgglse.argtypes = \
_libcula.culaDeviceCgglse.argtypes = [ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_void_p,
                                      ctypes.c_void_p]
def culaDeviceSgglse(m, n, p, a, lda, b, ldb, c, d, x):
    """
    Solve linear equality-constrained least squares problem.
    """

    status = _libcula.culaDeviceSgglse(m, n, p, int(a), lda, int(b),
                                       ldb, int(c), int(d), int(x))
    culaCheckStatus(status)

def culaDeviceCgglse(m, n, p, a, lda, b, ldb, c, d, x):
    """
    Solve linear equality-constrained least squares problem.
    """

    status = _libcula.culaDeviceCgglse(m, n, p, int(a), lda, int(b),
                                       ldb, int(c), int(d), int(x))
    culaCheckStatus(status)

# SGESVD, CGESVD
_libcula.culaDeviceSgesvd.restype = \
_libcula.culaDeviceCgesvd.restype = int
_libcula.culaDeviceSgesvd.argtypes = \
_libcula.culaDeviceCgesvd.argtypes = [ctypes.c_char,
                                      ctypes.c_char,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_void_p,
                                      ctypes.c_int,
                                      ctypes.c_void_p,
                                      ctypes.c_int]
def culaDeviceSgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt):
    """
    SVD decomposition.
    """

    jobu = jobu.encode('ascii')
    jobvt = jobvt.encode('ascii')
    status = _libcula.culaDeviceSgesvd(jobu, jobvt, m, n, int(a), lda,
                                       int(s), int(u), ldu, int(vt),
                                       ldvt)
    culaCheckStatus(status)

def culaDeviceCgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt):
    """
    SVD decomposition.
    """

    jobu = jobu.encode('ascii')
    jobvt = jobvt.encode('ascii')
    status = _libcula.culaDeviceCgesvd(jobu, jobvt, m, n, int(a), lda,
                                       int(s), int(u), ldu, int(vt),
                                       ldvt)
    culaCheckStatus(status)

# LAPACK functions available in CULA Dense:

# DGESV, ZGESV
try:
    _libcula.culaDeviceDgesv.restype = \
    _libcula.culaDeviceZgesv.restype = int
    _libcula.culaDeviceDgesv.argtypes = \
    _libcula.culaDeviceZgesv.argtypes = [ctypes.c_int,
                                         ctypes.c_int,
                                         ctypes.c_void_p,
                                         ctypes.c_int,
                                         ctypes.c_void_p,
                                         ctypes.c_void_p,
                                         ctypes.c_int]
except AttributeError:
    def culaDeviceDgesv(n, nrhs, a, lda, ipiv, b, ldb):
        """
        Solve linear system with LU factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgesv(n, nrhs, a, lda, ipiv, b, ldb):
        """
        Solve linear system with LU factorization.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceDgesv(n, nrhs, a, lda, ipiv, b, ldb):
        """
        Solve linear system with LU factorization.
        """

        status = _libcula.culaDeviceDgesv(n, nrhs, int(a), lda, int(ipiv),
                                          int(b), ldb)
        culaCheckStatus(status)

    def culaDeviceZgesv(n, nrhs, a, lda, ipiv, b, ldb):
        """
        Solve linear system with LU factorization.
        """

        status = _libcula.culaDeviceZgesv(n, nrhs, int(a), lda, int(ipiv),
                                          int(b), ldb)
        culaCheckStatus(status)

# DGETRF, ZGETRF
try:
    _libcula.culaDeviceDgetrf.restype = \
    _libcula.culaDeviceZgetrf.restype = int
    _libcula.culaDeviceDgetrf.argtypes = \
    _libcula.culaDeviceZgetrf.argtypes = [ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p]
except AttributeError:
    def culaDeviceDgetrf(m, n, a, lda, ipiv):
        """
        LU factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgetrf(m, n, a, lda, ipiv):
        """
        LU factorization.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceDgetrf(m, n, a, lda, ipiv):
        """
        LU factorization.
        """

        status = _libcula.culaDeviceDgetrf(m, n, int(a), lda, int(ipiv))
        culaCheckStatus(status)

    def culaDeviceZgetrf(m, n, a, lda, ipiv):
        """
        LU factorization.
        """

        status = _libcula.culaDeviceZgetrf(m, n, int(a), lda, int(ipiv))
        culaCheckStatus(status)


# DGEQRF, ZGEQRF
try:
    _libcula.culaDeviceDgeqrf.restype = \
    _libcula.culaDeviceZgeqrf.restype = int
    _libcula.culaDeviceDgeqrf.argtypes = \
    _libcula.culaDeviceZgeqrf.argtypes = [ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p]
except AttributeError:
    def culaDeviceDgeqrf(m, n, a, lda, tau):
        """
        QR factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeqrf(m, n, a, lda, tau):
        """
        QR factorization.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceDgeqrf(m, n, a, lda, tau):
        """
        QR factorization.
        """

        status = _libcula.culaDeviceDgeqrf(m, n, int(a), lda, int(tau))
        culaCheckStatus(status)

    def culaDeviceZgeqrf(m, n, a, lda, tau):
        """
        QR factorization.
        """

        status = _libcula.culaDeviceZgeqrf(m, n, int(a), lda, int(tau))
        culaCheckStatus(status)

# DORGQR, ZUNGQR
try:
    _libcula.culaDeviceDorgqr.restype = \
    _libcula.culaDeviceZungqr.restype = int
    _libcula.culaDeviceDorgqr.argtypes = \
    _libcula.culaDeviceZungqr.argtypes = [ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p]
except AttributeError:
    def culaDeviceDorgqr(m, n, k, a, lda, tau):
        """
        QR factorization - Generate Q from QR factorization
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDorgqr(m, n, k, a, lda, tau):
        """
        QR factorization - Generate Q from QR factorization
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceDorgqr(m, n, k, a, lda, tau):
        """
        QR factorization.
        """

        status = _libcula.culaDeviceDorgqr(m, n, k, int(a), lda, int(tau))
        culaCheckStatus(status)

    def culaDeviceZungqr(m, n, k, a, lda, tau):
        """
        QR factorization.
        """

        status = _libcula.culaDeviceZungqr(m, n, k, int(a), lda, int(tau))
        culaCheckStatus(status)

# SGETRI, CGETRI, DGETRI, ZGETRI
try:
    _libcula.culaDeviceSgetri.restype = \
    _libcula.culaDeviceCgetri.restype = \
    _libcula.culaDeviceDgetri.restype = \
    _libcula.culaDeviceZgetri.restype = int
    _libcula.culaDeviceSgetri.argtypes = \
    _libcula.culaDeviceCgetri.argtypes = \
    _libcula.culaDeviceDgetri.argtypes = \
    _libcula.culaDeviceZgetri.argtypes = [ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p]
except AttributeError:
    def culaDeviceSgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        status = _libcula.culaDeviceSgetri(n, int(a), lda, int(ipiv))
        culaCheckStatus(status)

    def culaDeviceCgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        status = _libcula.culaDeviceCgetri(n, int(a), lda, int(ipiv))
        culaCheckStatus(status)

    def culaDeviceDgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        status = _libcula.culaDeviceDgetri(n, int(a), lda, int(ipiv))
        culaCheckStatus(status)

    def culaDeviceZgetri(n, a, lda, ipiv):
        """
        Compute Inverse Matrix.
        """

        status = _libcula.culaDeviceZgetri(n, int(a), lda, int(ipiv))
        culaCheckStatus(status)


# DGELS, ZGELS
try:
    _libcula.culaDeviceDgels.restype = \
    _libcula.culaDeviceZgels.restype = int
    _libcula.culaDeviceDgels.argtypes = \
    _libcula.culaDeviceZgels.argtypes = [ctypes.c_char,
                                         ctypes.c_int,
                                         ctypes.c_int,
                                         ctypes.c_int,
                                         ctypes.c_void_p,
                                         ctypes.c_int,
                                         ctypes.c_void_p,
                                         ctypes.c_int]
except AttributeError:
    def culaDeviceDgels(trans, m, n, nrhs, a, lda, b, ldb):
        """
        Solve linear system with QR or LQ factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgels(trans, m, n, nrhs, a, lda, b, ldb):
        """
        Solve linear system with QR or LQ factorization.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceDgels(trans, m, n, nrhs, a, lda, b, ldb):
        """
        Solve linear system with QR or LQ factorization.
        """

        trans = trans.encode('ascii')
        status = _libcula.culaDeviceDgels(trans, m, n, nrhs, int(a),
                                          lda, int(b), ldb)
        culaCheckStatus(status)

    def culaDeviceZgels(trans, m, n, nrhs, a, lda, b, ldb):
        """
        Solve linear system with QR or LQ factorization.
        """

        trans = trans.encode('ascii')
        status = _libcula.culaDeviceZgels(trans, m, n, nrhs, int(a),
                                          lda, int(b), ldb)
        culaCheckStatus(status)

# DGGLSE, ZGGLSE
try:
    _libcula.culaDeviceDgglse.restype = \
    _libcula.culaDeviceZgglse.restype = int
    _libcula.culaDeviceDgglse.argtypes = \
    _libcula.culaDeviceZgglse.argtypes = [ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_void_p,
                                          ctypes.c_void_p]
except AttributeError:
    def culaDeviceDgglse(m, n, p, a, lda, b, ldb, c, d, x):
        """
        Solve linear equality-constrained least squares problem.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgglse(m, n, p, a, lda, b, ldb, c, d, x):
        """
        Solve linear equality-constrained least squares problem.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceDgglse(m, n, p, a, lda, b, ldb, c, d, x):
        """
        Solve linear equality-constrained least squares problem.
        """

        status = _libcula.culaDeviceDgglse(m, n, p, int(a), lda, int(b),
                                           ldb, int(c), int(d), int(x))
        culaCheckStatus(status)

    def culaDeviceZgglse(m, n, p, a, lda, b, ldb, c, d, x):
        """
        Solve linear equality-constrained least squares problem.
        """

        status = _libcula.culaDeviceZgglse(m, n, p, int(a), lda, int(b),
                                           ldb, int(c), int(d), int(x))
        culaCheckStatus(status)

# DGESVD, ZGESVD
try:
    _libcula.culaDeviceDgesvd.restype = \
    _libcula.culaDeviceZgesvd.restype = int
    _libcula.culaDeviceDgesvd.argtypes = \
    _libcula.culaDeviceZgesvd.argtypes = [ctypes.c_char,
                                          ctypes.c_char,
                                          ctypes.c_int,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_void_p,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int]
except AttributeError:
    def culaDeviceDgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt):
        """
        SVD decomposition.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt):
        """
        SVD decomposition.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceDgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt):
        """
        SVD decomposition.
        """

        jobu = jobu.encode('ascii')
        jobvt = jobvt.encode('ascii')
        status = _libcula.culaDeviceDgesvd(jobu, jobvt, m, n, int(a), lda,
                                           int(s), int(u), ldu, int(vt),
                                           ldvt)
        culaCheckStatus(status)

    def culaDeviceZgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt):
        """
        SVD decomposition.
        """

        jobu = jobu.encode('ascii')
        jobvt = jobvt.encode('ascii')
        status = _libcula.culaDeviceZgesvd(jobu, jobvt, m, n, int(a), lda,
                                           int(s), int(u), ldu, int(vt),
                                           ldvt)
        culaCheckStatus(status)

# SPOSV, CPOSV, DPOSV, ZPOSV
try:
    _libcula.culaDeviceSposv.restype = \
    _libcula.culaDeviceCposv.restype = \
    _libcula.culaDeviceDposv.restype = \
    _libcula.culaDeviceZposv.restype = int
    _libcula.culaDeviceSposv.argtypes = \
    _libcula.culaDeviceCposv.argtypes = \
    _libcula.culaDeviceDposv.argtypes = \
    _libcula.culaDeviceZposv.argtypes = [ctypes.c_char,
                                         ctypes.c_int,
                                         ctypes.c_int,
                                         ctypes.c_void_p,
                                         ctypes.c_int,
                                         ctypes.c_void_p,
                                         ctypes.c_int]
except AttributeError:
    def culaDeviceSposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        upio = upio.encode('ascii')
        status = _libcula.culaDeviceSposv(upio, n, nrhs, int(a), lda, int(b),
                                          ldb)
        culaCheckStatus(status)

    def culaDeviceCposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        upio = upio.encode('ascii')
        status = _libcula.culaDeviceCposv(upio, n, nrhs, int(a), lda, int(b),
                                          ldb)
        culaCheckStatus(status)

    def culaDeviceDposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        upio = upio.encode('ascii')
        status = _libcula.culaDeviceDposv(upio, n, nrhs, int(a), lda, int(b),
                                          ldb)
        culaCheckStatus(status)

    def culaDeviceZposv(upio, n, nrhs, a, lda, b, ldb):
        """
        Solve positive definite linear system with Cholesky factorization.
        """

        upio = upio.encode('ascii')
        status = _libcula.culaDeviceZposv(upio, n, nrhs, int(a), lda, int(b),
                                          ldb)
        culaCheckStatus(status)

# SPOTRF, CPOTRF, DPOTRF, ZPOTRF
try:
    _libcula.culaDeviceSpotrf.restype = \
    _libcula.culaDeviceCpotrf.restype = \
    _libcula.culaDeviceDpotrf.restype = \
    _libcula.culaDeviceZpotrf.restype = int
    _libcula.culaDeviceSpotrf.argtypes = \
    _libcula.culaDeviceCpotrf.argtypes = \
    _libcula.culaDeviceDpotrf.argtypes = \
    _libcula.culaDeviceZpotrf.argtypes = [ctypes.c_char,
                                          ctypes.c_int,
                                          ctypes.c_void_p,
                                          ctypes.c_int]
except AttributeError:
    def culaDeviceSpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceSpotrf(uplo, n, int(a), lda)
        culaCheckStatus(status)

    def culaDeviceCpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceCpotrf(uplo, n, int(a), lda)
        culaCheckStatus(status)

    def culaDeviceDpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceDpotrf(uplo, n, int(a), lda)
        culaCheckStatus(status)

    def culaDeviceZpotrf(uplo, n, a, lda):
        """
        Cholesky factorization.
        """

        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceZpotrf(uplo, n, int(a), lda)
        culaCheckStatus(status)

# SSYEV, DSYEV, CHEEV, ZHEEV
try:
    _libcula.culaDeviceSsyev.restype = \
    _libcula.culaDeviceDsyev.restype = \
    _libcula.culaDeviceCheev.restype = \
    _libcula.culaDeviceZheev.restype = int
    _libcula.culaDeviceSsyev.argtypes = \
    _libcula.culaDeviceDsyev.argtypes = \
    _libcula.culaDeviceCheev.argtypes = \
    _libcula.culaDeviceZheev.argtypes = [ctypes.c_char,
                                         ctypes.c_char,
                                         ctypes.c_int,
                                         ctypes.c_void_p,
                                         ctypes.c_int,
                                         ctypes.c_void_p]
except AttributeError:
    def culaDeviceSsyev(jobz, uplo, n, a, lda, w):
        """
        Symmetric eigenvalue decomposition.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDsyev(jobz, uplo, n, a, lda, w):
        """
        Symmetric eigenvalue decomposition.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCheev(jobz, uplo, n, a, lda, w):
        """
        Hermitian eigenvalue decomposition.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZheev(jobz, uplo, n, a, lda, w):
        """
        Hermitian eigenvalue decomposition.
        """

        raise NotImplementedError('CULA Dense required')
else:

    def culaDeviceSsyev(jobz, uplo, n, a, lda, w):
        """
        Symmetric eigenvalue decomposition.
        """

        jobz = jobz.encode('ascii')
        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceSsyev(jobz, uplo, n, int(a), lda, int(w))
        culaCheckStatus(status)

    def culaDeviceDsyev(jobz, uplo, n, a, lda, w):
        """
        Symmetric eigenvalue decomposition.
        """

        jobz = jobz.encode('ascii')
        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceDsyev(jobz, uplo, n, int(a), lda, int(w))
        culaCheckStatus(status)

    def culaDeviceCheev(jobz, uplo, n, a, lda, w):
        """
        Hermitian eigenvalue decomposition.
        """

        jobz = jobz.encode('ascii')
        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceCheev(jobz, uplo, n, int(a), lda, int(w))
        culaCheckStatus(status)

    def culaDeviceZheev(jobz, uplo, n, a, lda, w):
        """
        Hermitian eigenvalue decomposition.
        """

        jobz = jobz.encode('ascii')
        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceZheev(jobz, uplo, n, int(a), lda, int(w))
        culaCheckStatus(status)

# BLAS routines provided by CULA:

# SGEMM, DGEMM, CGEMM, ZGEMM
_libcula.culaDeviceSgemm.restype = \
_libcula.culaDeviceDgemm.restype = \
_libcula.culaDeviceCgemm.restype = \
_libcula.culaDeviceZgemm.restype = int

_libcula.culaDeviceSgemm.argtypes = [ctypes.c_char,
                                     ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_float,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_float,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

_libcula.culaDeviceDgemm.argtypes = [ctypes.c_char,
                                     ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_double,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_double,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

_libcula.culaDeviceCgemm.argtypes = [ctypes.c_char,
                                     ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     cuda.cuFloatComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     cuda.cuFloatComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

_libcula.culaDeviceZgemm.argtypes = [ctypes.c_char,
                                     ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     cuda.cuDoubleComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     cuda.cuDoubleComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

def culaDeviceSgemm(transa, transb, m, n, k, alpha, A, lda, B, ldb, beta, C, ldc):
    """
    Matrix-matrix product for general matrix.
    """

    transa = transa.encode('ascii')
    transb = transb.encode('ascii')
    status = _libcula.culaDeviceSgemm(transa, transb, m, n, k, alpha,
                           int(A), lda, int(B), ldb, beta, int(C), ldc)
    culaCheckStatus(status)

def culaDeviceDgemm(transa, transb, m, n, k, alpha, A, lda, B, ldb, beta, C, ldc):
    """
    Matrix-matrix product for general matrix.
    """

    transa = transa.encode('ascii')
    transb = transb.encode('ascii')
    status = _libcula.culaDeviceDgemm(transa, transb, m, n, k, alpha,
                           int(A), lda, int(B), ldb, beta, int(C), ldc)
    culaCheckStatus(status)

def culaDeviceCgemm(transa, transb, m, n, k, alpha, A, lda, B, ldb, beta, C, ldc):
    """
    Matrix-matrix product for complex general matrix.
    """

    transa = transa.encode('ascii')
    transb = transb.encode('ascii')
    status = _libcula.culaDeviceCgemm(transa, transb, m, n, k,
                                      cuda.cuFloatComplex(alpha.real,
                                                        alpha.imag),
                                      int(A), lda, int(B), ldb,
                                      cuda.cuFloatComplex(beta.real,
                                                        beta.imag),
                                      int(C), ldc)
    culaCheckStatus(status)

def culaDeviceZgemm(transa, transb, m, n, k, alpha, A, lda, B, ldb, beta, C, ldc):
    """
    Matrix-matrix product for complex general matrix.
    """

    transa = transa.encode('ascii')
    transb = transb.encode('ascii')
    status = _libcula.culaDeviceZgemm(transa, transb, m, n, k,
                                      cuda.cuDoubleComplex(alpha.real,
                                                        alpha.imag),
                                      int(A), lda, int(B), ldb,
                                      cuda.cuDoubleComplex(beta.real,
                                                        beta.imag),
                                      int(C), ldc)
    culaCheckStatus(status)

# SGEMV, DGEMV, CGEMV, ZGEMV
_libcula.culaDeviceSgemv.restype = \
_libcula.culaDeviceDgemv.restype = \
_libcula.culaDeviceCgemv.restype = \
_libcula.culaDeviceZgemv.restype = int

_libcula.culaDeviceSgemv.argtypes = [ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_float,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_float,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

_libcula.culaDeviceDgemv.argtypes = [ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_double,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_double,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

_libcula.culaDeviceCgemv.argtypes = [ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     cuda.cuFloatComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     cuda.cuFloatComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

_libcula.culaDeviceZgemv.argtypes = [ctypes.c_char,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     cuda.cuDoubleComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     ctypes.c_void_p,
                                     ctypes.c_int,
                                     cuda.cuDoubleComplex,
                                     ctypes.c_void_p,
                                     ctypes.c_int]

def culaDeviceSgemv(trans, m, n, alpha, A, lda, x, incx, beta, y, incy):
    """
    Matrix-vector product for real general matrix.
    """

    trans = trans.encode('ascii')
    status = _libcula.culaDeviceSgemv(trans, m, n, alpha, int(A), lda,
                           int(x), incx, beta, int(y), incy)
    culaCheckStatus(status)

def culaDeviceDgemv(trans, m, n, alpha, A, lda, x, incx, beta, y, incy):
    """
    Matrix-vector product for real general matrix.
    """

    trans = trans.encode('ascii')
    status = _libcula.culaDeviceDgemv(trans, m, n, alpha, int(A), lda,
                           int(x), incx, beta, int(y), incy)
    culaCheckStatus(status)


def culaDeviceCgemv(trans, m, n, alpha, A, lda, x, incx, beta, y, incy):
    """
    Matrix-vector product for complex general matrix.
    """

    trans = trans.encode('ascii')
    status = _libcula.culaDeviceCgemv(trans, m, n,
                           cuda.cuFloatComplex(alpha.real,
                                               alpha.imag),
                           int(A), lda, int(x), incx,
                           cuda.cuFloatComplex(beta.real,
                                               beta.imag),
                           int(y), incy)
    culaCheckStatus(status)

def culaDeviceZgemv(trans, m, n, alpha, A, lda, x, incx, beta, y, incy):
    """
    Matrix-vector product for complex general matrix.
    """

    trans = trans.encode('ascii')
    status = _libcula.culaDeviceZgemv(trans, m, n,
                           cuda.cuDoubleComplex(alpha.real,
                                               alpha.imag),
                           int(A), lda, int(x), incx,
                           cuda.cuDoubleComplex(beta.real,
                                               beta.imag),
                           int(y), incy)
    culaCheckStatus(status)
    
# SGEEV, DGEEV, CGEEV, ZGEEV
try:
    _libcula.culaDeviceSgeev.restype = \
    _libcula.culaDeviceDgeev.restype = \
    _libcula.culaDeviceCgeev.restype = \
    _libcula.culaDeviceZgeev.restype = int

    _libcula.culaDeviceSgeev.argtypes = \
    _libcula.culaDeviceDgeev.argtypes = [ctypes.c_char, #jobvl
                                         ctypes.c_char, #jobvr
                                         ctypes.c_int, #n,  the order of the matrix
                                         ctypes.c_void_p, #a
                                         ctypes.c_int, #lda
                                         ctypes.c_void_p, #wr
                                         ctypes.c_void_p, #wi
                                         ctypes.c_void_p, #vl
                                         ctypes.c_int, #ldvl
                                         ctypes.c_void_p, #vr
                                         ctypes.c_int] #ldvr
    _libcula.culaDeviceCgeev.argtypes = \
    _libcula.culaDeviceZgeev.argtypes = [ctypes.c_char, #jobvl
                                         ctypes.c_char, #jobvr
                                         ctypes.c_int, #n,  the order of the matrix
                                         ctypes.c_void_p, #a
                                         ctypes.c_int, #lda
                                         ctypes.c_void_p, #w
                                         ctypes.c_void_p, #vl
                                         ctypes.c_int, #ldvl
                                         ctypes.c_void_p, #vr
                                         ctypes.c_int] #ldvr
except AttributeError:
    def culaDeviceSgeev(jobvl, jobvr, n, a, lda, wr, wi, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDgeev(jobvl, jobvr, n, a, lda, wr, wi, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCgeev(jobvl, jobvr, n, a, lda, w, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeev(jobvl, jobvr, n, a, lda, w, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSgeev(jobvl, jobvr, n, a, lda, wr, wi, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        jobvl = jobvl.encode('ascii')
        jobvr = jobvr.encode('ascii')
        status = _libcula.culaDeviceSgeev(jobvl, jobvr, n, int(a), lda, int(wr), int(wi),
                               int(vl), ldvl, int(vr), ldvr)
        culaCheckStatus(status)

    def culaDeviceDgeev(jobvl, jobvr, n, a, lda, wr, wi, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        jobvl = jobvl.encode('ascii')
        jobvr = jobvr.encode('ascii')
        status = _libcula.culaDeviceDgeev(jobvl, jobvr, n, int(a), lda, int(wr), int(wi),
                               int(vl), ldvl, int(vr), ldvr)
        culaCheckStatus(status)

    def culaDeviceCgeev(jobvl, jobvr, n, a, lda, w, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        jobvl = jobvl.encode('ascii')
        jobvr = jobvr.encode('ascii')
        status = _libcula.culaDeviceCgeev(jobvl, jobvr, n, int(a), lda, int(w),
                               int(vl), ldvl, int(vr), ldvr)
        culaCheckStatus(status)

    def culaDeviceZgeev(jobvl, jobvr, n, a, lda, w, vl, ldvl, vr, ldvr):
        """
        General Eigenproblem solver.
        """

        jobvl = jobvl.encode('ascii')
        jobvr = jobvr.encode('ascii')
        status = _libcula.culaDeviceZgeev(jobvl, jobvr, n, int(a), lda, int(w),
                               int(vl), ldvl, int(vr), ldvr)
        culaCheckStatus(status)   
    
# Auxiliary routines:

try:
    _libcula.culaDeviceSgeTranspose.restype = \
    _libcula.culaDeviceDgeTranspose.restype = \
    _libcula.culaDeviceCgeTranspose.restype = \
    _libcula.culaDeviceZgeTranspose.restype = int
    _libcula.culaDeviceSgeTranspose.argtypes = \
    _libcula.culaDeviceDgeTranspose.argtypes = \
    _libcula.culaDeviceCgeTranspose.argtypes = \
    _libcula.culaDeviceZgeTranspose.argtypes = [ctypes.c_int,
                                                ctypes.c_int,
                                                ctypes.c_void_p,
                                                ctypes.c_int,
                                                ctypes.c_void_p,
                                                ctypes.c_int]
except AttributeError:
    def culaDeviceSgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of real general matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of real general matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of complex general matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of complex general matrix.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of real general matrix.
        """

        status = _libcula.culaDeviceSgeTranspose(m, n, int(A), lda, int(B), ldb)
        culaCheckStatus(status)

    def culaDeviceDgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of real general matrix.
        """

        status = _libcula.culaDeviceDgeTranspose(m, n, int(A), lda, int(B), ldb)
        culaCheckStatus(status)

    def culaDeviceCgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of complex general matrix.
        """

        status = _libcula.culaDeviceCgeTranspose(m, n, int(A), lda, int(B), ldb)
        culaCheckStatus(status)

    def culaDeviceZgeTranspose(m, n, A, lda, B, ldb):
        """
        Transpose of complex general matrix.
        """

        status = _libcula.culaDeviceZgeTranspose(m, n, int(A), lda, int(B), ldb)
        culaCheckStatus(status)

try:
    _libcula.culaDeviceSgeTransposeInplace.restype = \
    _libcula.culaDeviceDgeTransposeInplace.restype = \
    _libcula.culaDeviceCgeTransposeInplace.restype = \
    _libcula.culaDeviceZgeTransposeInplace.restype = int
    _libcula.culaDeviceSgeTransposeInplace.argtypes = \
    _libcula.culaDeviceDgeTransposeInplace.argtypes = \
    _libcula.culaDeviceCgeTransposeInplace.argtypes = \
    _libcula.culaDeviceZgeTransposeInplace.argtypes = [ctypes.c_int,
                                                    ctypes.c_void_p,
                                                    ctypes.c_int]
except AttributeError:
    def culaDeviceSgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of real square matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of real square matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of complex square matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of complex square matrix.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of real square matrix.
        """

        status = _libcula.culaDeviceSgeTransposeInplace(n, int(A), lda)
        culaCheckStatus(status)

    def culaDeviceDgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of real square matrix.
        """

        status = _libcula.culaDeviceDgeTransposeInplace(n, int(A), lda)
        culaCheckStatus(status)

    def culaDeviceCgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of complex square matrix.
        """

        status = _libcula.culaDeviceCgeTransposeInplace(n, int(A), lda)
        culaCheckStatus(status)

    def culaDeviceZgeTransposeInplace(n, A, lda):
        """
        Inplace transpose of complex square matrix.
        """

        status = _libcula.culaDeviceZgeTransposeInplace(n, int(A), lda)
        culaCheckStatus(status)

try:

    _libcula.culaDeviceCgeTransposeConjugate.restype = \
    _libcula.culaDeviceZgeTransposeConjugate.restype = int
    _libcula.culaDeviceCgeTransposeConjugate.argtypes = \
    _libcula.culaDeviceZgeTransposeConjugate.argtypes = [ctypes.c_int,
                                                        ctypes.c_int,
                                                        ctypes.c_void_p,
                                                        ctypes.c_int,
                                                        ctypes.c_void_p,
                                                        ctypes.c_int]
except AttributeError:
    def culaDeviceCgeTransposeConjugate(m, n, A, lda, B, ldb):
        """
        Conjugate transpose of complex general matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeTransposeConjugate(m, n, A, lda, B, ldb):
        """
        Conjugate transpose of complex general matrix.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceCgeTransposeConjugate(m, n, A, lda, B, ldb):
        """
        Conjugate transpose of complex general matrix.
        """

        status = _libcula.culaDeviceCgeTransposeConjugate(m, n, int(A), lda, int(B), ldb)
        culaCheckStatus(status)

    def culaDeviceZgeTransposeConjugate(m, n, A, lda, B, ldb):
        """
        Conjugate transpose of complex general matrix.
        """

        status = _libcula.culaDeviceZgeTransposeConjugate(m, n, int(A), lda, int(B), ldb)
        culaCheckStatus(status)

try:
    _libcula.culaDeviceCgeTransposeConjugateInplace.restype = \
    _libcula.culaDeviceZgeTransposeConjugateInplace.restype = int
    _libcula.culaDeviceCgeTransposeConjugateInplace.argtypes = \
    _libcula.culaDeviceZgeTransposeConjugateInplace.argtypes = [ctypes.c_int,
                                                                ctypes.c_void_p,
                                                                ctypes.c_int]
except AttributeError:
    def culaDeviceCgeTransposeConjugateInplace(n, A, lda):
        """
        Inplace conjugate transpose of complex square matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeTransposeConjugateInplace(n, A, lda):
        """
        Inplace conjugate transpose of complex square matrix.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceCgeTransposeConjugateInplace(n, A, lda):
        """
        Inplace conjugate transpose of complex square matrix.
        """

        status = _libcula.culaDeviceCgeTransposeConjugateInplace(n, int(A), lda)
        culaCheckStatus(status)

    def culaDeviceZgeTransposeConjugateInplace(n, A, lda):
        """
        Inplace conjugate transpose of complex square matrix.
        """

        status = _libcula.culaDeviceZgeTransposeConjugateInplace(n, int(A), lda)
        culaCheckStatus(status)

try:
    _libcula.culaDeviceCgeConjugate.restype = \
    _libcula.culaDeviceZgeConjugate.restype = int
    _libcula.culaDeviceCgeConjugate.argtypes = \
    _libcula.culaDeviceZgeConjugate.argtypes = [ctypes.c_int,
                                                ctypes.c_int,
                                                ctypes.c_void_p,
                                                ctypes.c_int]
except AttributeError:
    def culaDeviceCgeConjugate(m, n, A, lda):
        """
        Conjugate of complex general matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeConjugate(m, n, A, lda):
        """
        Conjugate of complex general matrix.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceCgeConjugate(m, n, A, lda):
        """
        Conjugate of complex general matrix.
        """

        status = _libcula.culaDeviceCgeConjugate(m, n, int(A), lda)
        culaCheckStatus(status)

    def culaDeviceZgeConjugate(m, n, A, lda):
        """
        Conjugate of complex general matrix.
        """

        status = _libcula.culaDeviceZgeConjugate(m, n, int(A), lda)
        culaCheckStatus(status)

try:
    _libcula.culaDeviceCtrConjugate.restype = \
    _libcula.culaDeviceZtrConjugate.restype = int
    _libcula.culaDeviceCtrConjugate.argtypes = \
    _libcula.culaDeviceZtrConjugate.argtypes = [ctypes.c_char,
                                                ctypes.c_char,
                                                ctypes.c_int,
                                                ctypes.c_int,
                                                ctypes.c_void_p,
                                                ctypes.c_int]
except AttributeError:
    def culaDeviceCtrConjugate(uplo, diag, m, n, A, lda):
        """
        Conjugate of complex upper or lower triangle matrix.
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZtrConjugate(uplo, diag, m, n, A, lda):
        """
        Conjugate of complex upper or lower triangle matrix.
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceCtrConjugate(uplo, diag, m, n, A, lda):
        """
        Conjugate of complex upper or lower triangle matrix.
        """

        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceCtrConjugate(uplo, diag, m, n, int(A), lda)
        culaCheckStatus(status)

    def culaDeviceZtrConjugate(uplo, diag, m, n, A, lda):
        """
        Conjugate of complex upper or lower triangle matrix.
        """

        uplo = uplo.encode('ascii')
        status = _libcula.culaDeviceZtrConjugate(uplo, diag, m, n, int(A), lda)
        culaCheckStatus(status)

try:
    _libcula.culaDeviceSgeNancheck.restype = \
    _libcula.culaDeviceDgeNancheck.restype = \
    _libcula.culaDeviceCgeNancheck.restype = \
    _libcula.culaDeviceZgeNancheck.restype = int
    _libcula.culaDeviceSgeNancheck.argtypes = \
    _libcula.culaDeviceDgeNancheck.argtypes = \
    _libcula.culaDeviceCgeNancheck.argtypes = \
    _libcula.culaDeviceZgeNancheck.argtypes = [ctypes.c_int,
                                                ctypes.c_int,
                                                ctypes.c_void_p,
                                                ctypes.c_int]
except AttributeError:
    def culaDeviceSgeNancheck(m, n, A, lda):
        """
        Check a real general matrix for invalid entries
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceDgeNancheck(m, n, A, lda):
        """
        Check a real general matrix for invalid entries
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceCgeNancheck(m, n, A, lda):
        """
        Check a complex general matrix for invalid entries
        """

        raise NotImplementedError('CULA Dense required')

    def culaDeviceZgeNancheck(m, n, A, lda):
        """
        Check a complex general matrix for invalid entries
        """

        raise NotImplementedError('CULA Dense required')
else:
    def culaDeviceSgeNancheck(m, n, A, lda):
        """
        Check a real general matrix for invalid entries
        """

        status = _libcula.culaDeviceSgeNancheck(m, n, int(A), lda)
        try:
            culaCheckStatus(status)
        except culaDataError:
            return True
        return False

    def culaDeviceDgeNancheck(m, n, A, lda):
        """
        Check a real general matrix for invalid entries
        """

        status = _libcula.culaDeviceDgeNancheck(m, n, int(A), lda)
        try:
            culaCheckStatus(status)
        except culaDataError:
            return True
        return False

    def culaDeviceCgeNancheck(m, n, A, lda):
        """
        Check a complex general matrix for invalid entries

        """

        status = _libcula.culaDeviceCgeNancheck(m, n, int(A), lda)
        try:
            culaCheckStatus(status)
        except culaDataError:
            return True
        return False

    def culaDeviceZgeNancheck(m, n, A, lda):
        """
        Check a complex general matrix for invalid entries
        """

        status = _libcula.culaDeviceZgeNancheck(m, n, int(A), lda)
        try:
            culaCheckStatus(status)
        except culaDataError:
            return True
        return False


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python

"""
Demonstrates how to use the PyCUDA interface to CUFFT to compute a
batch of 2D FFTs.
"""
from __future__ import print_function

import pycuda.autoinit
import pycuda.gpuarray as gpuarray
import numpy as np

import skcuda.fft as cu_fft

print('Testing fft/ifft..')
N = 256
batch_size = 16

x = np.empty((batch_size, N, N), np.float32)
xf = np.empty((batch_size, N, N), np.complex64)
y = np.empty((batch_size, N, N), np.float32)
for i in range(batch_size):
    x[i, :, :] = np.asarray(np.random.rand(N, N), np.float32)
    xf[i, :, :] = np.fft.fft2(x[i, :, :])
    y[i, :, :] = np.real(np.fft.ifft2(xf[i, :, :]))

x_gpu = gpuarray.to_gpu(x)
xf_gpu = gpuarray.empty((batch_size, N, N//2+1), np.complex64)
plan_forward = cu_fft.Plan((N, N), np.float32, np.complex64, batch_size)
cu_fft.fft(x_gpu, xf_gpu, plan_forward)

y_gpu = gpuarray.empty_like(x_gpu)
plan_inverse = cu_fft.Plan((N, N), np.complex64, np.float32, batch_size)
cu_fft.ifft(xf_gpu, y_gpu, plan_inverse, True)

print('Success status: %r' % np.allclose(y, y_gpu.get(), atol=1e-6))

print('Testing in-place fft..')
x = np.empty((batch_size, N, N), np.complex64)
x_gpu = gpuarray.to_gpu(x)

plan = cu_fft.Plan((N, N), np.complex64, np.complex64, batch_size)
cu_fft.fft(x_gpu, x_gpu, plan)

cu_fft.ifft(x_gpu, x_gpu, plan, True)

print('Success status: %r' % np.allclose(x, x_gpu.get(), atol=1e-6))



#!/usr/bin/env python

"""
Python interface to CUFFT functions.

Note: this module does not explicitly depend on PyCUDA.
"""

import ctypes
import operator
import re
import sys

# Load library:
_linux_version_list = [10.1, 10.0, 9.2, 9.1, 9.0, 8.0, 7.5, 7.0, 6.5, 6.0, 5.5, 5.0, 4.0]
_win32_version_list = [10, 100, 92, 91, 90, 80, 75, 70, 65, 60, 55, 50, 40]
if 'linux' in sys.platform:
    _libcufft_libname_list = ['libcufft.so'] + \
                             ['libcufft.so.%s' % v for v in _linux_version_list]
elif sys.platform == 'darwin':
    _libcufft_libname_list = ['libcufft.dylib']
elif sys.platform == 'win32':
    if sys.maxsize > 2**32:
        _libcufft_libname_list = ['cufft.dll'] + \
            ['cufft64_%s.dll' % v for v in _win32_version_list]
    else:
        _libcufft_libname_list = ['cufft.dll'] + \
            ['cufft32_%s.dll' % v for v in _win32_version_list]

else:
    raise RuntimeError('unsupported platform')

# Print understandable error message when library cannot be found:
_libcufft = None
for _libcufft_libname in _libcufft_libname_list:
    try:
        if sys.platform == 'win32':
            _libcufft = ctypes.windll.LoadLibrary(_libcufft_libname)
        else:
            _libcufft = ctypes.cdll.LoadLibrary(_libcufft_libname)
    except OSError:
        pass
    else:
        break
if _libcufft == None:
    raise OSError('cufft library not found')

# General CUFFT error:
class cufftError(Exception):
    """CUFFT error"""
    pass

# Exceptions corresponding to different CUFFT errors:
class cufftInvalidPlan(cufftError):
    """CUFFT was passed an invalid plan handle."""
    pass

class cufftAllocFailed(cufftError):
    """CUFFT failed to allocate GPU memory."""
    pass

class cufftInvalidType(cufftError):
    """The user requested an unsupported type."""
    pass

class cufftInvalidValue(cufftError):
    """The user specified a bad memory pointer."""
    pass

class cufftInternalError(cufftError):
    """Internal driver error."""
    pass

class cufftExecFailed(cufftError):
    """CUFFT failed to execute an FFT on the GPU."""
    pass

class cufftSetupFailed(cufftError):
    """The CUFFT library failed to initialize."""
    pass

class cufftInvalidSize(cufftError):
    """The user specified an unsupported FFT size."""
    pass

class cufftUnalignedData(cufftError):
    """Input or output does not satisfy texture alignment requirements."""
    pass

cufftExceptions = {
    0x1: cufftInvalidPlan,
    0x2: cufftAllocFailed,
    0x3: cufftInvalidType,
    0x4: cufftInvalidValue,
    0x5: cufftInternalError,
    0x6: cufftExecFailed,
    0x7: cufftSetupFailed,
    0x8: cufftInvalidSize,
    0x9: cufftUnalignedData
    }

class _types:
    """Some alias types."""
    plan = ctypes.c_int
    stream = ctypes.c_void_p
    worksize = ctypes.c_size_t

def cufftCheckStatus(status):
    """Raise an exception if the specified CUBLAS status is an error."""

    if status != 0:
        try:
            e = cufftExceptions[status]
        except KeyError:
            raise cufftError
        else:
            raise e

_libcufft.cufftGetVersion.restype = int
_libcufft.cufftGetVersion.argtypes = [ctypes.c_void_p]

def cufftGetVersion():
    """
    Get CUFFT version.
    """

    version = ctypes.c_int()
    result = _libcufft.cufftGetVersion(ctypes.byref(version))
    cufftCheckStatus(result)
    return version.value

_cufft_version = int(cufftGetVersion())

class _cufft_version_req(object):
    """
    Decorator to replace function with a placeholder that raises an exception
    if a specified condition on the installed CUFFT version `v` is not satisfied.
    """

    def __init__(self, v, op):
        self.op_str = op
        if op == '>':
            self.op = operator.gt
        elif op == '>=':
            self.op = operator.ge
        elif op == '==':
            self.op = operator.eq
        elif op == '<':
            self.op = operator.lt
        elif op == '<=':
            self.op = operator.le
        else:
            raise ValueError('unrecognized comparison operator')
        self.vs = str(v)
        if isinstance(v, int):
            self.vi = str(v)
            if len(self.vi) != 4:
                raise ValueError('integer version number must be 4 digits')
        else:
            major, minor = re.search(r'(\d+)\.(\d+)', self.vs).groups()
            self.vi = major.ljust(len(major)+1, '0')+minor.ljust(2, '0')

    def __call__(self,f):
        def f_new(*args,**kwargs):
            raise NotImplementedError('CUFFT '+self.op_str+' '+self.vs+' required')
        f_new.__doc__ = f.__doc__

        if self.op(_cufft_version, int(self.vi)):
            return f
        else:
            return f_new

# Data transformation types:
CUFFT_R2C = 0x2a
CUFFT_C2R = 0x2c
CUFFT_C2C = 0x29
CUFFT_D2Z = 0x6a
CUFFT_Z2D = 0x6c
CUFFT_Z2Z = 0x69

# Transformation directions:
CUFFT_FORWARD = -1
CUFFT_INVERSE = 1

# FFTW compatibility modes:
CUFFT_COMPATIBILITY_NATIVE = 0x00
CUFFT_COMPATIBILITY_FFTW_PADDING = 0x01
CUFFT_COMPATIBILITY_FFTW_ASYMMETRIC = 0x02
CUFFT_COMPATIBILITY_FFTW_ALL = 0x03

# FFT functions implemented by CUFFT:
_libcufft.cufftPlan1d.restype = int
_libcufft.cufftPlan1d.argtypes = [ctypes.c_void_p,
                                  ctypes.c_int,
                                  ctypes.c_int,
                                  ctypes.c_int]
def cufftPlan1d(nx, fft_type, batch):
    """
    Create 1D FFT plan configuration.

    References
    ----------
    `cufftPlan1d <http://docs.nvidia.com/cuda/cufft/#function-cufftplan1d>`_
    """

    plan = _types.plan()
    status = _libcufft.cufftPlan1d(ctypes.byref(plan), nx, fft_type, batch)
    cufftCheckStatus(status)
    return plan

_libcufft.cufftPlan2d.restype = int
_libcufft.cufftPlan2d.argtypes = [ctypes.c_void_p,
                                  ctypes.c_int,
                                  ctypes.c_int,
                                  ctypes.c_int]
def cufftPlan2d(nx, ny, fft_type):
    """
    Create 2D FFT plan configuration.

    References
    ----------
    `cufftPlan2d <http://docs.nvidia.com/cuda/cufft/#function-cufftplan2d>`_
    """

    plan = _types.plan()
    status = _libcufft.cufftPlan2d(ctypes.byref(plan), nx, ny,
                                   fft_type)
    cufftCheckStatus(status)
    return plan

_libcufft.cufftPlan3d.restype = int
_libcufft.cufftPlan3d.argtypes = [ctypes.c_void_p,
                                  ctypes.c_int,
                                  ctypes.c_int,
                                  ctypes.c_int,
                                  ctypes.c_int]
def cufftPlan3d(nx, ny, nz, fft_type):
    """
    Create 3D FFT plan configuration.

    References
    ----------
    `cufftPlan3d <http://docs.nvidia.com/cuda/cufft/#function-cufftplan3d>`_
    """

    plan = _types.plan()
    status = _libcufft.cufftPlan3d(ctypes.byref(plan), nx, ny, nz,
                                   fft_type)
    cufftCheckStatus(status)
    return plan

_libcufft.cufftPlanMany.restype = int
_libcufft.cufftPlanMany.argtypes = [ctypes.c_void_p,
                                    ctypes.c_int,
                                    ctypes.c_void_p,
                                    ctypes.c_void_p,
                                    ctypes.c_int,
                                    ctypes.c_int,
                                    ctypes.c_void_p,
                                    ctypes.c_int,
                                    ctypes.c_int,
                                    ctypes.c_int,
                                    ctypes.c_int]
def cufftPlanMany(rank, n,
                  inembed, istride, idist,
                  onembed, ostride, odist, fft_type, batch):
    """
    Create batched FFT plan configuration.

    References
    ----------
    `cufftPlanMany <http://docs.nvidia.com/cuda/cufft/#function-cufftplanmany>`_
    """

    plan = _types.plan()
    status = _libcufft.cufftPlanMany(ctypes.byref(plan), rank, n,
                                     inembed, istride, idist,
                                     onembed, ostride, odist,
                                     fft_type, batch)
    cufftCheckStatus(status)
    return plan

_libcufft.cufftDestroy.restype = int
_libcufft.cufftDestroy.argtypes = [_types.plan]
def cufftDestroy(plan):
    """
    Destroy FFT plan.

    References
    ----------
    `cufftDestroy <http://docs.nvidia.com/cuda/cufft/#function-cufftdestroy>`_
    """

    status = _libcufft.cufftDestroy(plan)
    cufftCheckStatus(status)

if _cufft_version <= 9010:
    _libcufft.cufftSetCompatibilityMode.restype = int
    _libcufft.cufftSetCompatibilityMode.argtypes = [_types.plan,
                                                    ctypes.c_int]
@_cufft_version_req(9.1, '<=')
def cufftSetCompatibilityMode(plan, mode):
    """
    Set FFTW compatibility mode.

    References
    ----------
    `cufftSetCompatibilityMode <http://docs.nvidia.com/cuda/cufft/#function-cufftsetcompatibilitymode>`_
    """

    status = _libcufft.cufftSetCompatibilityMode(plan, mode)
    cufftCheckStatus(status)

_libcufft.cufftExecC2C.restype = int
_libcufft.cufftExecC2C.argtypes = [_types.plan,
                                   ctypes.c_void_p,
                                   ctypes.c_void_p,
                                   ctypes.c_int]
def cufftExecC2C(plan, idata, odata, direction):
    """Execute single precision complex-to-complex transform plan as
    specified by `direction`.

    References
    ----------
    `cufftExecC2C <http://docs.nvidia.com/cuda/cufft/#function-cufftexecc2c-cufftexecz2z>`_
    """

    status = _libcufft.cufftExecC2C(plan, idata, odata,
                                    direction)
    cufftCheckStatus(status)

_libcufft.cufftExecR2C.restype = int
_libcufft.cufftExecR2C.argtypes = [_types.plan,
                                   ctypes.c_void_p,
                                   ctypes.c_void_p]
def cufftExecR2C(plan, idata, odata):
    """
    Execute single precision real-to-complex forward transform plan.

    References
    ----------
    `cufftExecR2C <http://docs.nvidia.com/cuda/cufft/#function-cufftexecr2c-cufftexecd2z>`_
    """

    status = _libcufft.cufftExecR2C(plan, idata, odata)
    cufftCheckStatus(status)

_libcufft.cufftExecC2R.restype = int
_libcufft.cufftExecC2R.argtypes = [_types.plan,
                                   ctypes.c_void_p,
                                   ctypes.c_void_p]
def cufftExecC2R(plan, idata, odata):
    """
    Execute single precision complex-to-real reverse transform plan.

    References
    ----------
    `cufftExecC2R <http://docs.nvidia.com/cuda/cufft/#function-cufftexecc2r-cufftexecz2d>`_
    """

    status = _libcufft.cufftExecC2R(plan, idata, odata)
    cufftCheckStatus(status)

_libcufft.cufftExecZ2Z.restype = int
_libcufft.cufftExecZ2Z.argtypes = [_types.plan,
                                   ctypes.c_void_p,
                                   ctypes.c_void_p,
                                   ctypes.c_int]
def cufftExecZ2Z(plan, idata, odata, direction):
    """
    Execute double precision complex-to-complex transform plan as
    specified by `direction`.

    References
    ----------
    `cufftExecZ2Z <http://docs.nvidia.com/cuda/cufft/#function-cufftexecc2c-cufftexecz2z>`_
"""

    status = _libcufft.cufftExecZ2Z(plan, idata, odata,
                                    direction)
    cufftCheckStatus(status)

_libcufft.cufftExecD2Z.restype = int
_libcufft.cufftExecD2Z.argtypes = [_types.plan,
                                   ctypes.c_void_p,
                                   ctypes.c_void_p]
def cufftExecD2Z(plan, idata, odata):
    """
    Execute double precision real-to-complex forward transform plan.

    References
    ----------
    `cufftExecD2Z <http://docs.nvidia.com/cuda/cufft/#function-cufftexecr2c-cufftexecd2z>`_
    """

    status = _libcufft.cufftExecD2Z(plan, idata, odata)
    cufftCheckStatus(status)

_libcufft.cufftExecZ2D.restype = int
_libcufft.cufftExecZ2D.argtypes = [_types.plan,
                                   ctypes.c_void_p,
                                   ctypes.c_void_p]
def cufftExecZ2D(plan, idata, odata):
    """
    Execute double precision complex-to-real transform plan.

    References
    ----------
    `cufftExecZ2D <http://docs.nvidia.com/cuda/cufft/#function-cufftexecc2r-cufftexecz2d>`_
    """

    status = _libcufft.cufftExecZ2D(plan, idata, odata)
    cufftCheckStatus(status)

_libcufft.cufftSetStream.restype = int
_libcufft.cufftSetStream.argtypes = [_types.plan,
                                     _types.stream]
def cufftSetStream(plan, stream):
    """
    Associate a CUDA stream with a CUFFT plan.

    References
    ----------
    `cufftSetStream <http://docs.nvidia.com/cuda/cufft/#function-cufftsetstream>`_
    """

    status = _libcufft.cufftSetStream(plan, stream)
    cufftCheckStatus(status)

_libcufft.cufftEstimate1d.restype = int
_libcufft.cufftEstimate1d.argtypes = [ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def cufftEstimate1d(nx, fft_type, batch=1):
    """
    Return estimated work area for 1D FFT.

    References
    ----------
    `cufftEstimate1d <http://docs.nvidia.com/cuda/cufft/#function-cufftestimate1d>`_
    """

    worksize = _types.worksize()
    status = _libcufft.cufftEstimate1d(nx, fft_type, batch,
                                       ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftEstimate2d.restype = int
_libcufft.cufftEstimate2d.argtypes = [ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def cufftEstimate2d(nx, ny, fft_type):
    """
    Return estimated work area for 2D FFT.

    References
    ----------
    `cufftEstimate2d <http://docs.nvidia.com/cuda/cufft/#function-cufftestimate2d>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftEstimate2d(nx, ny, fft_type,
                                       ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftEstimate3d.restype = int
_libcufft.cufftEstimate3d.argtypes = [ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def cufftEstimate3d(nx, ny, nz, fft_type):
    """
    Return estimated work area for 3D FFT.

    References
    ----------
    `cufftEstimate3d <http://docs.nvidia.com/cuda/cufft/#function-cufftestimate3d>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftEstimate3d(nx, ny, nz, fft_type,
                                       ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftEstimateMany.restype = int
_libcufft.cufftEstimateMany.argtypes = [ctypes.c_int,
                                        ctypes.c_void_p,
                                        ctypes.c_void_p,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_void_p,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_void_p]
def cufftEstimateMany(rank, n,
                      inembed, istride, idist,
                      onembed, ostride, odist, fft_type, batch):
    """
    Return estimated work area for batched FFT.

    References
    ----------
    `cufftEstimateMany <http://docs.nvidia.com/cuda/cufft/#function-cufftestimatemany>`_
    """

    worksize = _types.worksize()
    status = _libcufft.cufftEstimateMany(rank, n,
                                         inembed, istride, idist,
                                         onembed, ostride, odist,
                                         fft_type, batch,
                                         ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftGetSize1d.restype = int
_libcufft.cufftGetSize1d.argtypes = [_types.plan,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_void_p]
def cufftGetSize1d(plan, nx, fft_type, batch=1):
    """
    Return more accurate estimate of work area size required for 1D FFT,
    taking into account any plan settings that may have been made.

    References
    ----------
    `cufftGetSize1d <http://docs.nvidia.com/cuda/cufft/#function-cufftgetsize1d>`_
    """

    worksize = _types.worksize()
    status = _libcufft.cufftGetSize1d(plan, nx, fft_type, batch,
                                      ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftGetSize2d.restype = int
_libcufft.cufftGetSize2d.argtypes = [_types.plan,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_void_p]
def cufftGetSize2d(plan, nx, ny, fft_type):
    """
    Return more accurate estimate of work area size required for 2D FFT,
    taking into account any plan settings that may have been made.

    References
    ----------
    `cufftGetSize2d <http://docs.nvidia.com/cuda/cufft/#function-cufftgetsize2d>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftGetSize2d(plan, nx, ny, fft_type,
                                      ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftGetSize3d.restype = int
_libcufft.cufftGetSize3d.argtypes = [_types.plan,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_int,
                                     ctypes.c_void_p]
def cufftGetSize3d(plan, nx, ny, nz, fft_type):
    """
    Return more accurate estimate of work area size required for 3D FFT,
    taking into account any plan settings that may have been made.

    References
    ----------
    `cufftGetSize3d <http://docs.nvidia.com/cuda/cufft/#function-cufftgetsize3d>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftGetSize3d(plan, nx, ny, nz, fft_type,
                                      ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftGetSizeMany.restype = int
_libcufft.cufftGetSizeMany.argtypes = [_types.plan,
                                       ctypes.c_int,
                                       ctypes.c_void_p,
                                       ctypes.c_void_p,
                                       ctypes.c_int,
                                       ctypes.c_int,
                                       ctypes.c_void_p,
                                       ctypes.c_int,
                                       ctypes.c_int,
                                       ctypes.c_int,
                                       ctypes.c_int,
                                       ctypes.c_void_p]
def cufftGetSizeMany(plan, rank, n,
                     inembed, istride, idist,
                     onembed, ostride, odist, fft_type, batch):
    """
    Return more accurate estimate of work area size required for batched FFT,
    taking into account any plan settings that may have been made.

    References
    ----------
    `cufftGetSizeMany <http://docs.nvidia.com/cuda/cufft/#function-cufftgetsizemany>`_
    """

    worksize = _types.worksize()
    status = _libcufft.cufftGetSizeMany(plan, rank, n,
                                        inembed, istride, idist,
                                        onembed, ostride, odist,
                                        fft_type, batch,
                                        ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftGetSize.restype = int
_libcufft.cufftGetSize.argtypes = [_types.plan,
                                   ctypes.c_void_p]
def cufftGetSize(plan):
    """
    Return actual size of work area for FFT described in plan.

    References
    ----------
    `cufftGetSize <http://docs.nvidia.com/cuda/cufft/#function-cufftgetsize>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftGetSize(plan, ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftCreate.restype = int
_libcufft.cufftCreate.argtypes = [ctypes.c_void_p]
def cufftCreate():
    """
    Create FFT plan handle.

    References
    ----------
    `cufftCreate <http://docs.nvidia.com/cuda/cufft/#function-cufftcreate>`_
    """
    plan = _types.plan()
    status = _libcufft.cufftCreate(ctypes.byref(plan))
    cufftCheckStatus(status)
    return plan

_libcufft.cufftMakePlan1d.restype = int
_libcufft.cufftMakePlan1d.argtypes = [_types.plan,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def cufftMakePlan1d(plan, nx, fft_type, batch):
    """
    Create 1D FFT plan configuration.

    References
    ----------
    `cufftMakePlan1d <http://docs.nvidia.com/cuda/cufft/#function-cufftmakeplan1d>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftMakePlan1d(plan, nx, fft_type, batch,
                                       ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftMakePlan2d.restype = int
_libcufft.cufftMakePlan2d.argtypes = [_types.plan,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def cufftMakePlan2d(plan, nx, ny, fft_type):
    """
    Create 2D FFT plan configuration.

    References
    ----------
    `cufftMakePlan2d <http://docs.nvidia.com/cuda/cufft/#function-cufftmakeplan2d>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftMakePlan2d(plan, nx, ny, fft_type,
                                       ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftMakePlan3d.restype = int
_libcufft.cufftMakePlan3d.argtypes = [_types.plan,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_void_p]
def cufftMakePlan3d(plan, nx, ny, nz, fft_type):
    """
    Create 3D FFT plan configuration.

    References
    ----------
    `cufftMakePlan3d <http://docs.nvidia.com/cuda/cufft/#function-cufftmakeplan3d>`_
    """
    worksize = _types.worksize()
    status = _libcufft.cufftMakePlan3d(plan, nx, ny, nz, fft_type,
                                       ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftMakePlanMany.restype = int
_libcufft.cufftMakePlanMany.argtypes = [_types.plan,
                                        ctypes.c_int,
                                        ctypes.c_void_p,
                                        ctypes.c_void_p,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_void_p,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        ctypes.c_void_p]
def cufftMakePlanMany(plan, rank, n,
                      inembed, istride, idist,
                      onembed, ostride, odist, fft_type, batch):
    """
    Create batched FFT plan configuration.

    References
    ----------
    `cufftMakePlanMany <http://docs.nvidia.com/cuda/cufft/#function-cufftmakeplanmany>`_
    """

    worksize = _types.worksize()
    status = _libcufft.cufftMakePlanMany(plan, rank, n,
                                         inembed, istride, idist,
                                         onembed, ostride, odist,
                                         fft_type, batch,
                                         ctypes.byref(worksize))
    cufftCheckStatus(status)
    return worksize.value

_libcufft.cufftSetAutoAllocation.restype = int
_libcufft.cufftSetAutoAllocation.argtypes = [_types.plan,
                                             ctypes.c_int]
def cufftSetAutoAllocation(plan, auto_allocate):
    """
    Indicate whether the caller intends to allocate and manage work areas for
    plans that have been generated.

    References
    ----------
    `cufftSetAutoAllocation <http://docs.nvidia.com/cuda/cufft/#function-cufftsetautoallocation>`_
    """
    status = _libcufft.cufftSetAutoAllocation(plan, auto_allocate)
    cufftCheckStatus(status)

_libcufft.cufftSetWorkArea.restype = int
_libcufft.cufftSetWorkArea.argtypes = [_types.plan,
                                       ctypes.c_void_p]
def cufftSetWorkArea(plan, work_area):
    """
    Override the work area pointer associated with a plan.

    References
    ----------
    `cufftSetworkArea <http://docs.nvidia.com/cuda/cufft/#function-cufftsetworkarea>`_
    """
    status = _libcufft.cufftSetWorkArea(plan, work_area)
    cufftCheckStatus(status)



#!/usr/bin/env python

"""
Unit tests for skcuda.special
"""


from unittest import main, makeSuite, TestCase, TestSuite

import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
from pycuda.tools import clear_context_caches, make_default_context
import numpy as np
import scipy as sp
import scipy.special
import skcuda.linalg as linalg
import skcuda.misc as misc
import skcuda.special as special

drv.init()

class test_special(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ctx = make_default_context()

    @classmethod
    def tearDownClass(cls):
        cls.ctx.pop()
        clear_context_caches()

    def setUp(self):
        np.random.seed(0)
        linalg.init()

    def test_sici_float32(self):
        x = np.array([[1, 2], [3, 4]], np.float32)
        x_gpu = gpuarray.to_gpu(x)
        (si_gpu, ci_gpu) = special.sici(x_gpu)
        (si, ci) = scipy.special.sici(x)
        np.testing.assert_allclose(si, si_gpu.get(), atol=1e-7)
        np.testing.assert_allclose(ci, ci_gpu.get(), atol=1e-7)

    def test_sici_float64(self):
        x = np.array([[1, 2], [3, 4]], np.float64)
        x_gpu = gpuarray.to_gpu(x)
        (si_gpu, ci_gpu) = special.sici(x_gpu)
        (si, ci) = scipy.special.sici(x)
        np.testing.assert_allclose(si, si_gpu.get())
        np.testing.assert_allclose(ci, ci_gpu.get())

    def test_exp1_complex64(self):
        z = np.asarray(np.random.rand(4, 4) + 1j*np.random.rand(4, 4), np.complex64)
        z_gpu = gpuarray.to_gpu(z)
        e_gpu = special.exp1(z_gpu)
        np.testing.assert_allclose(sp.special.exp1(z), e_gpu.get())

    def test_exp1_complex128(self):
        z = np.asarray(np.random.rand(4, 4) + 1j*np.random.rand(4, 4), np.complex128)
        z_gpu = gpuarray.to_gpu(z)
        e_gpu = special.exp1(z_gpu)
        np.testing.assert_allclose(sp.special.exp1(z), e_gpu.get())

    def test_expi_complex64(self):
        z = np.asarray(np.random.rand(4, 4) + 1j*np.random.rand(4, 4), np.complex64)
        z_gpu = gpuarray.to_gpu(z)
        e_gpu = special.expi(z_gpu)
        np.testing.assert_allclose(sp.special.expi(z), e_gpu.get())

    def test_expi_complex128(self):
        z = np.asarray(np.random.rand(4, 4) + 1j*np.random.rand(4, 4), np.complex128)
        z_gpu = gpuarray.to_gpu(z)
        e_gpu = special.expi(z_gpu)
        np.testing.assert_allclose(sp.special.expi(z), e_gpu.get())

def suite():
    context = make_default_context()
    device = context.get_device()
    context.pop()

    s = TestSuite()
    s.addTest(test_special('test_sici_float32'))
    s.addTest(test_special('test_exp1_complex64'))
    s.addTest(test_special('test_expi_complex64'))
    if misc.get_compute_capability(device) >= 1.3:
        s.addTest(test_special('test_sici_float64'))
        s.addTest(test_special('test_exp1_complex128'))
        s.addTest(test_special('test_expi_complex128'))
    return s

if __name__ == '__main__':
    main(defaultTest = 'suite')



#!/usr/bin/env python

"""
Utility functions.
"""

import sys
import ctypes.util
import os
import re
import subprocess
import struct

import sys
if sys.version_info < (3,):
    range = xrange

try:
    import elftools
except ImportError:
    import re


    def get_soname(filename):
        """
        Retrieve SONAME of shared library.

        Parameters
        ----------
        filename : str
            Full path to shared library.

        Returns
        -------
        soname : str
            SONAME of shared library.

        Notes
        -----
        This function uses the `objdump` system command on Linux and
        'otool' on MacOS (Darwin).
        """

        if sys.platform == 'darwin':
            cmds = ['otool', '-L', filename]
        else:
            # Fallback to linux... what about windows?
            cmds = ['objdump', '-p', filename]

        try:
            p = subprocess.Popen(cmds, stdout=subprocess.PIPE,
                                 env=dict(os.environ, LANG="en"))
            out = p.communicate()[0].decode()
        except:
            raise RuntimeError('error executing {0}'.format(cmds))

        if sys.platform == 'darwin':
            result = re.search('^\s@rpath/(lib.+.dylib)', out, re.MULTILINE)
        else:
            result = re.search('^\s+SONAME\s+(.+)$',out,re.MULTILINE)

        if result:
            return result.group(1)
        else:
            # No SONAME found:
            raise RuntimeError('no library name found for {0}'.format(
                (filename,)))

else:
    import ctypes
    import elftools.elf.elffile as elffile
    import elftools.construct.macros as macros
    import elftools.elf.structs as structs

    def get_soname(filename):
        """
        Retrieve SONAME of shared library.

        Parameters
        ----------
        filename : str
            Full path to shared library.

        Returns
        -------
        soname : str
            SONAME of shared library.

        Notes
        -----
        This function uses the pyelftools [ELF] package.

        References
        ----------
        .. [ELF] http://pypi.python.org/pypi/pyelftools

        """

        stream = open(filename, 'rb')
        f = elffile.ELFFile(stream)
        dynamic = f.get_section_by_name('.dynamic')
        dynstr = f.get_section_by_name('.dynstr')

        # Handle libraries built for different machine architectures:
        if f.header['e_machine'] == 'EM_X86_64':
            st = structs.Struct('Elf64_Dyn',
                                macros.ULInt64('d_tag'),
                                macros.ULInt64('d_val'))
        elif f.header['e_machine'] == 'EM_386':
            st = structs.Struct('Elf32_Dyn',
                                macros.ULInt32('d_tag'),
                                macros.ULInt32('d_val'))
        else:
            raise RuntimeError('unsupported machine architecture')

        entsize = dynamic['sh_entsize']
        for k in range(dynamic['sh_size']/entsize):
            result = st.parse(dynamic.data()[k*entsize:(k+1)*entsize])

            # The following value for the SONAME tag is specified in elf.h:
            if result.d_tag == 14:
                return dynstr.get_string(result.d_val)

        # No SONAME found:
        return ''

def find_lib_path(name):
    """
    Find full path of a shared library.

    Searches for the full path of a shared library. On Posix operating systems 
    other than MacOS, this function checks the directories listed in 
    LD_LIBRARY_PATH (if any) and in the ld.so cache. 

    Parameter
    ---------
    name : str
        Link name of library, e.g., cublas for libcublas.so.*.

    Returns
    -------
    path : str
        Full path to library.

    Notes
    -----
    Code adapted from ctypes.util module. Doesn't check whether the
    architectures of libraries found in LD_LIBRARY_PATH directories conform
    to that of the machine.
    """

    if sys.platform == 'win32':
        return ctypes.util.find_library(name)

    # MacOS has no ldconfig:
    if sys.platform == 'darwin':
        from ctypes.macholib.dyld import dyld_find as _dyld_find
        possible = ['lib%s.dylib' % name,
                    '%s.dylib' % name,
                    '%s.framework/%s' % (name, name)]
        for name in possible:
            try:
                return _dyld_find(name)
            except ValueError:
                continue
        return None

    # First, check the directories in LD_LIBRARY_PATH:
    expr = r'\s+(lib%s\.[^\s]+)\s+\-\>' % re.escape(name)
    for dir_path in filter(len,
            os.environ.get('LD_LIBRARY_PATH', '').split(':')):
        f = os.popen('/sbin/ldconfig -Nnv %s 2>/dev/null' % dir_path)
        try:
            data = f.read()
        finally:
            f.close()
        res = re.search(expr, data)
        if res:
            return os.path.join(dir_path, res.group(1))

    # Next, check the ld.so cache:
    uname = os.uname()[4]
    if uname.startswith("arm"):
        uname = "arm"
    if struct.calcsize('l') == 4:
        machine = uname + '-32'
    else:
        machine = uname + '-64'
    mach_map = {
        'x86_64-64': 'libc6,x86-64',
        'ppc64-64': 'libc6,64bit',
        'sparc64-64': 'libc6,64bit',
        's390x-64': 'libc6,64bit',
        'ia64-64': 'libc6,IA-64',
        'arm-32': 'libc6(,hard-float)?',
        }
    abi_type = mach_map.get(machine, 'libc6')
    expr = r'\s+lib%s\.[^\s]+\s+\(%s.*\=\>\s(.+)' % (re.escape(name), abi_type)
    f = os.popen('/sbin/ldconfig -p 2>/dev/null')
    try:
        data = f.read()
    finally:
        f.close()
    res = re.search(expr, data)
    if not res:
        return None
    return res.group(1)



"""
Unit tests for skcuda.integrate
"""

from unittest import main, TestCase, TestSuite

import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
from pycuda.tools import clear_context_caches, make_default_context
import numpy as np
import skcuda.misc as misc
import skcuda.integrate as integrate
import scipy.integrate

drv.init()

class test_integrate(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ctx = make_default_context()
        integrate.init()

    @classmethod
    def tearDownClass(cls):
        integrate.shutdown()
        cls.ctx.pop()
        clear_context_caches()

    def setUp(self):
        np.random.seed(0)

    def test_trapz_float32(self):
        x = np.asarray(np.random.rand(10), np.float32)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz(x_gpu)
        np.testing.assert_allclose(np.trapz(x), z)

    def test_trapz_float64(self):
        x = np.asarray(np.random.rand(10), np.float64)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz(x_gpu)
        np.testing.assert_allclose(np.trapz(x), z)

    def test_trapz_complex64(self):
        x = np.asarray(np.random.rand(10)+1j*np.random.rand(10), np.complex64)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz(x_gpu)
        np.testing.assert_allclose(np.trapz(x), z)

    def test_trapz_complex128(self):
        x = np.asarray(np.random.rand(10)+1j*np.random.rand(10), np.complex128)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz(x_gpu)
        np.testing.assert_allclose(np.trapz(x), z)

    def test_simps_float32(self):
        x = np.asarray(np.random.rand(10), np.float32)
        x_gpu = gpuarray.to_gpu(x)
        np.testing.assert_allclose(scipy.integrate.simps(x),
                        integrate.simps(x_gpu))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='first'),
                        integrate.simps(x_gpu, even='first'))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='last'),
                        integrate.simps(x_gpu, even='last'))

    def test_simps_float64(self):
        x = np.asarray(np.random.rand(10), np.float64)
        x_gpu = gpuarray.to_gpu(x)
        np.testing.assert_allclose(scipy.integrate.simps(x),
                        integrate.simps(x_gpu))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='first'),
                        integrate.simps(x_gpu, even='first'))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='last'),
                        integrate.simps(x_gpu, even='last'))

    def test_simps_complex64(self):
        x = np.asarray(np.random.rand(10)+1j*np.random.rand(10), np.complex64)
        x_gpu = gpuarray.to_gpu(x)
        np.testing.assert_allclose(scipy.integrate.simps(x),
                        integrate.simps(x_gpu))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='first'),
                        integrate.simps(x_gpu, even='first'))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='last'),
                        integrate.simps(x_gpu, even='last'))

    def test_simps_complex128(self):
        x = np.asarray(np.random.rand(10)+1j*np.random.rand(10), np.complex128)
        x_gpu = gpuarray.to_gpu(x)
        np.testing.assert_allclose(scipy.integrate.simps(x),
                        integrate.simps(x_gpu))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='first'),
                        integrate.simps(x_gpu, even='first'))
        np.testing.assert_allclose(scipy.integrate.simps(x, even='last'),
                        integrate.simps(x_gpu, even='last'))

    def test_trapz2d_float32(self):
        x = np.asarray(np.random.rand(5, 5), np.float32)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz2d(x_gpu)
        np.testing.assert_allclose(np.trapz(np.trapz(x)), z)

    def test_trapz2d_float64(self):
        x = np.asarray(np.random.rand(5, 5), np.float64)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz2d(x_gpu)
        np.testing.assert_allclose(np.trapz(np.trapz(x)), z)

    def test_trapz2d_complex64(self):
        x = np.asarray(np.random.rand(5, 5)+1j*np.random.rand(5, 5), np.complex64)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz2d(x_gpu)
        np.testing.assert_allclose(np.trapz(np.trapz(x)), z)

    def test_trapz2d_complex128(self):
        x = np.asarray(np.random.rand(5, 5)+1j*np.random.rand(5, 5), np.complex128)
        x_gpu = gpuarray.to_gpu(x)
        z = integrate.trapz2d(x_gpu)
        np.testing.assert_allclose(np.trapz(np.trapz(x)), z)

def suite():
    context = make_default_context()
    device = context.get_device()
    context.pop()

    s = TestSuite()
    s.addTest(test_integrate('test_trapz_float32'))
    s.addTest(test_integrate('test_trapz_complex64'))
    s.addTest(test_integrate('test_simps_float32'))
    s.addTest(test_integrate('test_simps_complex64'))
    s.addTest(test_integrate('test_trapz2d_float32'))
    s.addTest(test_integrate('test_trapz2d_complex64'))
    if misc.get_compute_capability(device) >= 1.3:
        s.addTest(test_integrate('test_trapz_float64'))
        s.addTest(test_integrate('test_trapz_complex128'))
        s.addTest(test_integrate('test_simps_float64'))
        s.addTest(test_integrate('test_simps_complex128'))
        s.addTest(test_integrate('test_trapz2d_float64'))
        s.addTest(test_integrate('test_trapz2d_complex128'))
    return s

if __name__ == '__main__':
    main(defaultTest = 'suite')



#!/usr/bin/env python

"""
PyCUDA-based FFT functions.
"""

import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
import pycuda.elementwise as el
from pycuda.tools import context_dependent_memoize
import pycuda.tools as tools
import numpy as np

from . import cufft
from .cufft import CUFFT_COMPATIBILITY_NATIVE, \
     CUFFT_COMPATIBILITY_FFTW_PADDING, \
     CUFFT_COMPATIBILITY_FFTW_ASYMMETRIC, \
     CUFFT_COMPATIBILITY_FFTW_ALL
from . import cudart
from . import misc

class Plan:
    """
    CUFFT plan class.

    This class represents an FFT plan for CUFFT.

    Parameters
    ----------
    shape : tuple of ints
        Transform shape. May contain more than 3 elements.
    in_dtype : { numpy.float32, numpy.float64, numpy.complex64, numpy.complex128 }
        Type of input data.
    out_dtype : { numpy.float32, numpy.float64, numpy.complex64, numpy.complex128 }
        Type of output data.
    batch : int
        Number of FFTs to configure in parallel (default is 1).
    stream : pycuda.driver.Stream
        Stream with which to associate the plan. If no stream is specified,
        the default stream is used.
    mode : int
        FFTW compatibility mode. Ignored in CUDA 9.2 and later.
    inembed : numpy.array with dtype=numpy.int32
        number of elements in each dimension of the input array
    istride : int
        distance between two successive input elements in the least significant
        (innermost) dimension
    idist : int
        distance between the first element of two consective batches in the
        input data
    onembed : numpy.array with dtype=numpy.int32
        number of elements in each dimension of the output array
    ostride : int
        distance between two successive output elements in the least significant
        (innermost) dimension
    odist : int
        distance between the first element of two consective batches in the
        output data
    auto_allocate : bool
        indicates whether the caller intends to allocate and manage the work area
    """

    def __init__(self, shape, in_dtype, out_dtype, batch=1, stream=None,
                 mode=0x01, inembed=None, istride=1, idist=0, onembed=None,
                 ostride=1, odist=0, auto_allocate=True):

        if np.isscalar(shape):
            self.shape = (shape, )
        else:
            self.shape = shape

        self.in_dtype = in_dtype
        self.out_dtype = out_dtype

        if batch <= 0:
            raise ValueError('batch size must be greater than 0')
        self.batch = batch

        # Determine type of transformation:
        if in_dtype == np.float32 and out_dtype == np.complex64:
            self.fft_type = cufft.CUFFT_R2C
            self.fft_func = cufft.cufftExecR2C
        elif in_dtype == np.complex64 and out_dtype == np.float32:
            self.fft_type = cufft.CUFFT_C2R
            self.fft_func = cufft.cufftExecC2R
        elif in_dtype == np.complex64 and out_dtype == np.complex64:
            self.fft_type = cufft.CUFFT_C2C
            self.fft_func = cufft.cufftExecC2C
        elif in_dtype == np.float64 and out_dtype == np.complex128:
            self.fft_type = cufft.CUFFT_D2Z
            self.fft_func = cufft.cufftExecD2Z
        elif in_dtype == np.complex128 and out_dtype == np.float64:
            self.fft_type = cufft.CUFFT_Z2D
            self.fft_func = cufft.cufftExecZ2D
        elif in_dtype == np.complex128 and out_dtype == np.complex128:
            self.fft_type = cufft.CUFFT_Z2Z
            self.fft_func = cufft.cufftExecZ2Z
        else:
            raise ValueError('unsupported input/output type combination')

        # Check for double precision support:
        capability = misc.get_compute_capability(misc.get_current_device())
        if capability < 1.3 and \
           (misc.isdoubletype(in_dtype) or misc.isdoubletype(out_dtype)):
            raise RuntimeError('double precision requires compute capability '
                               '>= 1.3 (you have %g)' % capability)

        if inembed is not None:
            inembed = inembed.ctypes.data
        if onembed is not None:
            onembed = onembed.ctypes.data

        # Set up plan:
        if len(self.shape) <= 0:
            raise ValueError('invalid transform size')
        n = np.asarray(self.shape, np.int32)
        self.handle = cufft.cufftCreate()

        # Set FFTW compatibility mode:
        if cufft._cufft_version <= 9010:
            cufft.cufftSetCompatibilityMode(self.handle, mode)

        # Set auto-allocate mode
        cufft.cufftSetAutoAllocation(self.handle, auto_allocate)
        self.worksize = cufft.cufftMakePlanMany(
            self.handle, len(self.shape), n.ctypes.data, inembed, istride, idist,
            onembed, ostride, odist, self.fft_type, self.batch)

        # Associate stream with plan:
        if stream != None:
            cufft.cufftSetStream(self.handle, stream.handle)

    def set_work_area(self, work_area):
        """
        Associate a caller-managed work area with the plan.

        Parameters
        ----------
        work_area : pycuda.gpuarray.GPUArray
        """
        cufft.cufftSetWorkArea(self.handle, int(work_area.gpudata))

    def __del__(self):

        # Don't complain if handle destruction fails because the plan
        # may have already been cleaned up:
        try:
            cufft.cufftDestroy(self.handle)
        except:
            pass

@context_dependent_memoize
def _get_scale_kernel(dtype):
    ctype = tools.dtype_to_ctype(dtype)
    return el.ElementwiseKernel(
        "{ctype} scale, {ctype} *x".format(ctype=ctype),
        "x[i] /= scale")

def _fft(x_gpu, y_gpu, plan, direction, scale=None):
    """
    Fast Fourier Transform.

    Parameters
    ----------
    x_gpu : pycuda.gpuarray.GPUArray
        Input array.
    y_gpu : pycuda.gpuarray.GPUArray
        Output array.
    plan : Plan
        FFT plan.
    direction : { cufft.CUFFT_FORWARD, cufft.CUFFT_INVERSE }
        Transform direction. Only affects in-place transforms.

    Optional Parameters
    -------------------
    scale : int or float
        Scale the values in the output array by dividing them by this value.

    Notes
    -----
    This function should not be called directly.
    """

    if (x_gpu.gpudata == y_gpu.gpudata) and \
           plan.fft_type not in [cufft.CUFFT_C2C, cufft.CUFFT_Z2Z]:
        raise ValueError('can only compute in-place transform of complex data')

    if direction == cufft.CUFFT_FORWARD and \
           plan.in_dtype in np.sctypes['complex'] and \
           plan.out_dtype in np.sctypes['float']:
        raise ValueError('cannot compute forward complex -> real transform')

    if direction == cufft.CUFFT_INVERSE and \
           plan.in_dtype in np.sctypes['float'] and \
           plan.out_dtype in np.sctypes['complex']:
        raise ValueError('cannot compute inverse real -> complex transform')

    if plan.fft_type in [cufft.CUFFT_C2C, cufft.CUFFT_Z2Z]:
        plan.fft_func(plan.handle, int(x_gpu.gpudata), int(y_gpu.gpudata),
                      direction)
    else:
        plan.fft_func(plan.handle, int(x_gpu.gpudata),
                      int(y_gpu.gpudata))

    # Scale the result by dividing it by the number of elements:
    if scale is not None:
        func = _get_scale_kernel(y_gpu.dtype)
        func(y_gpu.dtype.type(scale), y_gpu)

def fft(x_gpu, y_gpu, plan, scale=False):
    """
    Fast Fourier Transform.

    Compute the FFT of some data in device memory using the
    specified plan.

    Parameters
    ----------
    x_gpu : pycuda.gpuarray.GPUArray
        Input array.
    y_gpu : pycuda.gpuarray.GPUArray
        FFT of input array.
    plan : Plan
        FFT plan.
    scale : bool, optional
        If True, scale the computed FFT by the number of elements in
        the input array.

    Examples
    --------
    >>> import pycuda.autoinit
    >>> import pycuda.gpuarray as gpuarray
    >>> import numpy as np
    >>> from skcuda.fft import fft, Plan
    >>> N = 128
    >>> x = np.asarray(np.random.rand(N), np.float32)
    >>> xf = np.fft.fft(x)
    >>> x_gpu = gpuarray.to_gpu(x)
    >>> xf_gpu = gpuarray.empty(N/2+1, np.complex64)
    >>> plan = Plan(x.shape, np.float32, np.complex64)
    >>> fft(x_gpu, xf_gpu, plan)
    >>> np.allclose(xf[0:N/2+1], xf_gpu.get(), atol=1e-6)
    True

    Returns
    -------
    y_gpu : pycuda.gpuarray.GPUArray
        Computed FFT.

    Notes
    -----
    For real to complex transformations, this function computes
    N/2+1 non-redundant coefficients of a length-N input signal.
    """

    if scale == True:
        _fft(x_gpu, y_gpu, plan, cufft.CUFFT_FORWARD, x_gpu.size/plan.batch)
    else:
        _fft(x_gpu, y_gpu, plan, cufft.CUFFT_FORWARD)

def ifft(x_gpu, y_gpu, plan, scale=False):
    """
    Inverse Fast Fourier Transform.

    Compute the inverse FFT of some data in device memory using the
    specified plan.

    Parameters
    ----------
    x_gpu : pycuda.gpuarray.GPUArray
        Input array.
    y_gpu : pycuda.gpuarray.GPUArray
        Inverse FFT of input array.
    plan : Plan
        FFT plan.
    scale : bool, optional
        If True, scale the computed inverse FFT by the number of
        elements in the output array.

    Examples
    --------
    >>> import pycuda.autoinit
    >>> import pycuda.gpuarray as gpuarray
    >>> import numpy as np
    >>> from skcuda.fft import fft, Plan
    >>> N = 128
    >>> x = np.asarray(np.random.rand(N), np.float32)
    >>> xf = np.asarray(np.fft.fft(x), np.complex64)
    >>> xf_gpu = gpuarray.to_gpu(xf[0:N/2+1])
    >>> x_gpu = gpuarray.empty(N, np.float32)
    >>> plan = Plan(N, np.complex64, np.float32)
    >>> ifft(xf_gpu, x_gpu, plan, True)
    >>> np.allclose(x, x_gpu.get(), atol=1e-6)
    True

    Notes
    -----
    For complex to real transformations, this function assumes the
    input contains N/2+1 non-redundant FFT coefficents of a signal of
    length N.
    """

    if scale == True:
        _fft(x_gpu, y_gpu, plan, cufft.CUFFT_INVERSE, y_gpu.size/plan.batch)
    else:
        _fft(x_gpu, y_gpu, plan, cufft.CUFFT_INVERSE)

if __name__ == "__main__":
    import doctest
    doctest.testmod()
